import greenfoot.*;
import java.io.*;
import java.net.*;
import java.util.*;
import java.awt.Color;

/**
 * Where the gameplay happens.
 * 
 * @author Diljot Garcha
 * @since 2015-05-11
 */
public class Game
{
    //Instance variables
    //Boolean variables that define the current state of the game
    private static boolean gameStarted;
    private static boolean gameStopped;
    
    //Strings and Integers that represent the player and computer moves
    private String playerMove;
    private int playerMoveVal;
    private String compMove;
    private int compMoveVal;
    
    //Arrays and ArrayLists that hold player moves in different ways
    private ArrayList <Integer> listOfPlayerMoves;
    private ArrayList <Integer> rollingListPlayerMoves;
    private ArrayList <Integer> shortListPlayerMoves;
    private int[] permList = {
        0,
        1,
        2,
        0,
        1,
        2,
        -1
    };
    
    //Integers for statistics
    private int numOfPlayerWins;
    private int numOfPlayerLosses;
    private int numOfTies;
    private int numOfGames;
    
    //Integers that the smartAI() uses to "adapt".
    private int index;
    private int start;
    private int tempNumWins;
    private int tempNumTies;
    private int tempNumLosses;
    private boolean movePredicted;
    private int numOfRepetitionPredictions;
    private boolean movePredictedRepetition;
    private boolean continueRepetitionPredictions;
    
    //Random class - super awesome!
    Random myRand;
    
    /**
     * Default constructor for instances of the Game class.
     */
    public Game() 
    {
        //Initialize variables
        gameStarted = false;
        gameStopped = false;
        continueRepetitionPredictions = true;
        
        //Initialize ArrayLists and our Random class
        listOfPlayerMoves = new ArrayList <Integer>();
        rollingListPlayerMoves = new ArrayList <Integer>();
        shortListPlayerMoves = new ArrayList <Integer>();
        myRand = new Random();
        
        //Fill up the various player moves ArrayLists with some default values.
        for(int i = 0; i<3; i++)
        {
            shortListPlayerMoves.add(-1);
        }
        //Fill up the listOfPlayerMoves and the rollingListPlayerMoves ArrayLists with an extremely common RPS pattern.
        for(int move:permList)
        {
            listOfPlayerMoves.add(move);
            rollingListPlayerMoves.add(move);
        }
        //Load the approx 1000 RPS player moves that were generated during the development of this game.
        //We load these files from a web resource, then if the file is not found or some other error occurs, we load it from the backup location.
        try
        {
            loadPlayerMoves(new URL("https://files.app.net/7jzk1_X66")); //Load from main location
        }
        catch(IOException e)
        {
            if(e instanceof MalformedURLException)
            {
                //Print the error and continue execution without the data.
                System.out.println(e);
                System.out.println("Error caught—continuing execution without data.");
            }
            else if(e instanceof FileNotFoundException)
            {
                try
                {
                    loadPlayerMoves(new URL("https://dl.dropboxusercontent.com/u/43117301/RPS_DSG_playerMoves.txt")); //Load from backup location
                }
                catch(IOException ex)
                {
                    //Print the error and continue execution without the data.
                    System.out.println(ex);
                    System.out.println("Error caught—continuing execution without data.");
                }
            }
        }
    }
    
    /**
     * The method that will start the game.
     * @param move string that contains "ROCK", "PAPER", or "SCISSORS". Using the constants in the Playground class are recommended.
     */
    public void startGame(String move)
    {
        //If the String is not "ROCK", "PAPER", or "SCISSORS", then throw an IllegalArgumentException.
        if(!(move.equals(Playground.ROCK)) && !(move.equals(Playground.PAPER)) && !(move.equals(Playground.SCISSORS)))
        {
            throw new IllegalArgumentException("\nUnknown player move: \"" + move + "\", must be \"ROCK\", \"PAPER\", or \"SCISSORS\".");
        }
        gameStarted = true; //Start the game
        playerMove = move;  //Store the move in playerMove.
        //Identify the move value
        if(playerMove.equals(Playground.ROCK))
        {
            playerMoveVal = 0;
        }
        else if(playerMove.equals(Playground.PAPER))
        {
            playerMoveVal = 1;
        }
        else if(playerMove.equals(Playground.SCISSORS))
        {
            playerMoveVal = 2;
        }
        listOfPlayerMoves.add(playerMoveVal); //Add the move to an ArrayList
        continueGame(); //Helper method to contunue the game
    }

    /**
     * Helper method to continue the game, offers an easy way to see what is happening procedurally.
     */
    private void continueGame()
    {
        smartAI(); //Call the smartAI()
        animateGame(); //Helper method to animate the game
        checkForWinner(); //Helper method to check for a winner
        updateScoreboard(); //Helper method to update the scoreboard
        finishGame(); //Helper method to finish the game
    }
    
    /**
     * A basic algorithm to make a random choice.
     */
    private void basicAI()
    {
        //Generate a random number
        double randomNum = (Math.random() * 3);
        int roundedRandomNum = (int)(randomNum + 0.5); //Round the random number
        int castedRandomNum = (int)(randomNum); //Cast the random number
        if(roundedRandomNum == 3) //If the number is out of bounds (not 0, 1, or 2), decrement it.
        {
            roundedRandomNum--;
        }
        int averageNum = ((roundedRandomNum + castedRandomNum) / 2); //Take the average of the 2 numbers
        //Use that number to make our move
        if(averageNum == 0)
        {
            compMove = Playground.ROCK;
        }
        else if(averageNum == 1)
        {
            compMove = Playground.PAPER;
        }
        else if(averageNum == 2)
        {
            compMove = Playground.SCISSORS;
        }
        compMoveVal = averageNum; //Assign the number to a instance variable
    }
    
    /**
     * An algorithm designed to predict the player's next move.
     */
    private void smartAI()
    {
        //Initialize and create some variables
        index = 0;
        movePredicted = false;
        movePredictedRepetition = false;
        int predictedPlayerMove = -1;
        
        //For loop that runs the length of the rollingListPlayerMoves ArrayList.
        for(int i = start; i < rollingListPlayerMoves.size(); i++)
        {
            //Check if the player's 3rd last move and their ith move in the history of player moves are equal.
            if(shortListPlayerMoves.get(0) == rollingListPlayerMoves.get(i))
            {
                //Check to make sure we do not go out of bounds.
                if((i + 1) == rollingListPlayerMoves.size())
                {
                    basicAI();
                    return;
                }
                //Check if the player's 2nd last move and their i+1th move in the history of player moves are equal.
                if(shortListPlayerMoves.get(1) == rollingListPlayerMoves.get(i+1))
                {
                    //Check if the player's 3rd last move and their 2nd last move are equal
                    //Additionally check if myRand.nextBoolean() is true and if continueRepetitionPredictions is true.
                    if((rollingListPlayerMoves.get(i) == rollingListPlayerMoves.get(i+1)) && myRand.nextBoolean() && continueRepetitionPredictions)
                    {
                        if(myRand.nextBoolean())
                        {
                            //Assume the player will change their move to something that would beat your move, assuming you play what would beat their repeated move.
                            if(rollingListPlayerMoves.get(i+1) == 0)
                            {
                                predictedPlayerMove = 2;
                            }
                            else if(rollingListPlayerMoves.get(i+1) == 1)
                            {
                                predictedPlayerMove = 0;
                            }
                            else if(rollingListPlayerMoves.get(i+1) == 2)
                            {
                                predictedPlayerMove = 1;
                            }
                            //Increment and change the value of some variables.
                            numOfRepetitionPredictions++;
                            movePredictedRepetition = true;
                            TextBox myMessage = new TextBox(".", 24, 0, 48, Color.CYAN, Playground.BACKGROUND_COLOR);
                            break;
                        }
                        else
                        {
                            //Assume the player will not change their move, and will keep playing their repeated move.
                            //Increment and change the value of some variables.
                            numOfRepetitionPredictions++;
                            movePredictedRepetition = true;
                            predictedPlayerMove = rollingListPlayerMoves.get(i+1);
                            TextBox myMessage = new TextBox(".", 24, 0, 48, Color.MAGENTA, Playground.BACKGROUND_COLOR);
                            break;
                        }
                    }
                    //Check to make sure we do not go out of bounds.
                    if((i + 2) == rollingListPlayerMoves.size())
                    {
                        basicAI();
                        return;
                    }
                    //Check if the player's 3rd last move and their i+2nd move in the history of player moves are equal.
                    if(shortListPlayerMoves.get(2) == rollingListPlayerMoves.get(i+2))
                    {
                        //Check to make sure we do not go out of bounds.
                        if((i+3) == rollingListPlayerMoves.size())
                        {
                            basicAI();
                            return;
                        }
                        //The pattern has been found!
                        //Predict the player's next move as what they played when they last played this pattern.
                        index = i;
                        movePredicted = true;
                        predictedPlayerMove = rollingListPlayerMoves.get(i+3);
                        TextBox myMessage = new TextBox(".", 24, 0, 48, Color.YELLOW, Playground.BACKGROUND_COLOR);
                        break;
                    }
                }
            }
        }
        
        //If the predictedPlayerMove Integer is still -1, then revert to the basicAI();
        //Otherwise make our move based on their predicted move
        if(predictedPlayerMove == -1)
        {
            basicAI();
            return;
        }
        else if(predictedPlayerMove == 0)
        {
            compMove = Playground.PAPER;
            compMoveVal = 1;
        }
        else if(predictedPlayerMove == 1)
        {
            compMove = Playground.SCISSORS;
            compMoveVal = 2;
        }
        else if(predictedPlayerMove == 2)
        {
            compMove = Playground.ROCK;
            compMoveVal = 0;
        }
    }
    
    /**
     * A helper method that will animate the game.
     */
    private void animateGame()
    {
        //The animation for "ROCK", "PAPER", and "SCISSORS".
        for(int i = 0; i<3; i++)
        {
            Playground.world.addObject(new Rock(), 4, 6);
            Playground.world.addObject(new Rock(true), 20, 6);
            if(i == 0)
            {
                TextBox myMessage = new TextBox("\"ROCK\"", 12, 13, 48, Color.BLACK, Playground.BACKGROUND_COLOR);
            }
            else if(i == 1)
            {
                TextBox myMessage = new TextBox("\"PAPER\"", 12, 13, 48, Color.BLACK, Playground.BACKGROUND_COLOR);
            }
            else if(i == 2)
            {
                TextBox myMessage = new TextBox("\"SCISSORS\"", 12, 13, 48, Color.BLACK, Playground.BACKGROUND_COLOR);
            }
            Greenfoot.delay(50);
            Playground.world.removeObjects(Playground.world.getObjects(Rock.class));
            Playground.world.removeObjects(Playground.world.getObjects(TextBox.class));
            Greenfoot.delay(50);
        }
        
        //Reveal the player's move.
        if(playerMoveVal == 0)
        {
            Playground.world.addObject(new Rock(), 4, 6);
        }
        else if(playerMoveVal == 1)
        {
            Playground.world.addObject(new Paper(), 4, 6);
        }
        else if(playerMoveVal == 2)
        {
            Playground.world.addObject(new Scissors(), 4, 6);
        }
        TextBox playerMoveMessage = new TextBox(playerMove, 4, 11, 48, Color.BLACK, Playground.BACKGROUND_COLOR);
        
        //Reveal the computers's move.
        if(compMoveVal == 0)
        {
            Playground.world.addObject(new Rock(true), 20, 6);
        }
        else if(compMoveVal == 1)
        {
            Playground.world.addObject(new Paper(true), 20, 6);
        }
        else if(compMoveVal == 2)
        {
            Playground.world.addObject(new Scissors(true), 20, 6);
        }
        TextBox compMoveMessage = new TextBox(compMove, 20, 11, 48, Color.BLACK, Playground.BACKGROUND_COLOR);
    }
    
    /**
     * A helper method that will check for a winner.
     * After it checks for a winner, it will increment some variables and then adapt the AI if needed.
     */
    private void checkForWinner()
    {
        if((playerMoveVal == 2) && (playerMoveVal > compMoveVal) && (compMoveVal != 0))
        {
            TextBox myMessage = new TextBox("You have won!", 12, 13, 48, Color.BLUE, Playground.BACKGROUND_COLOR);
            numOfPlayerWins++;
            tempNumWins++;
            tempNumTies = 0;
            tempNumLosses = 0;
            sanitizeList();
        }
        else if((playerMoveVal == 1) && (playerMoveVal > compMoveVal) && (compMoveVal != 2))
        {
            TextBox myMessage = new TextBox("You have won!", 12, 13, 48, Color.BLUE, Playground.BACKGROUND_COLOR);
            numOfPlayerWins++;
            tempNumWins++;
            tempNumTies = 0;
            tempNumLosses = 0;
            sanitizeList();
        }
        else if((playerMoveVal == 0) && (playerMoveVal < compMoveVal) && (compMoveVal == 2))
        {
            TextBox myMessage = new TextBox("You have won!", 12, 13, 48, Color.BLUE, Playground.BACKGROUND_COLOR);
            numOfPlayerWins++;
            tempNumWins++;
            tempNumTies = 0;
            tempNumLosses = 0;
            sanitizeList();
        }
        else if(playerMoveVal == compMoveVal)
        {
            TextBox myMessage = new TextBox("It's a tie!", 12, 13, 48, Color.BLACK, Playground.BACKGROUND_COLOR);
            numOfTies++;
            tempNumTies++;
            tempNumWins = 0;
            tempNumLosses = 0;
            sanitizeList();
        }
        else
        {
            TextBox myMessage = new TextBox("You have lost!", 12, 13, 48, Color.RED, Playground.BACKGROUND_COLOR);
            numOfPlayerLosses++;
            tempNumLosses++;
            tempNumWins = 0;
            tempNumTies = 0;
        }
    }
    
    /**
     * A helper method that will update the scoreboard.
     */
    private void updateScoreboard()
    {
        //Remove all score objects
        Playground.world.removeObjects(Playground.world.getObjects(Score.class));
        
        //Make new score objects
        Score myWins = new Score("" + numOfPlayerWins, 9, 3, 48, Color.BLUE, Playground.BACKGROUND_COLOR);
        Score myTies = new Score("" + numOfTies, 12, 3, 48, Color.BLACK, Playground.BACKGROUND_COLOR);
        Score myLosses = new Score("" + numOfPlayerLosses, 15, 3, 48, Color.RED, Playground.BACKGROUND_COLOR);  
    }
    
    /**
     * A helper method that will finish the game.
     */
    private void finishGame()
    {
        //Update the shortListPlayerMoves ArrayList that keeps track of the player's last three moves.
        rollingListPlayerMoves.add(playerMoveVal);
        if(listOfPlayerMoves.size() >= 3)
        {
            for(int i = 1; i <= shortListPlayerMoves.size(); i++)
            {
                shortListPlayerMoves.set((i-1), listOfPlayerMoves.get(listOfPlayerMoves.size() - i));
            }
        }
        Collections.reverse(shortListPlayerMoves);
        
        //Update some more variables
        numOfGames = (numOfPlayerWins+numOfTies+numOfPlayerLosses);
        if(tempNumLosses >= 3)
        {
            start = 0;
        }
        if(tempNumWins >= 3 || tempNumTies >= 3)
        {
            start = 1000;
        }
        
        //Make 100 games the maximum number of games
        if(numOfGames == 100)
        {
            stopGame();
            return;
        }
        restart(); //If the user has not played 100 games or has not stopped the game, Restart the game.
    }
    
    /**
     * A method to "sanitize" the rollingListPlayerMoves ArrayList.
     * This method helps the smartAI() adapt
     */
    private void sanitizeList()
    {
        //If the move has been predicted and the AI fails to win, then "sanitize" the list.
        if(movePredicted)
        {
            //Set some values to -1 in the rollingListPlayerMoves ArrayList
            rollingListPlayerMoves.set(index, -1);
            //Check to make sure we do not go out of bounds.
            if((index + 1) >= rollingListPlayerMoves.size())
            {
                return;
            }
            //Set some values to -1 in the rollingListPlayerMoves ArrayList
            rollingListPlayerMoves.set(index+1, -1);
            //Check to make sure we do not go out of bounds.
            if((index + 2) >= rollingListPlayerMoves.size())
            {
                return;
            }
            //Set some values to -1 in the rollingListPlayerMoves ArrayList
            rollingListPlayerMoves.set(index+2, -1);
            //Check to make sure we do not go out of bounds.
            if((index + 3) >= rollingListPlayerMoves.size())
            {
                return;
            }
            //Set some values to -1 in the rollingListPlayerMoves ArrayList
            rollingListPlayerMoves.set(index+3, -1);
        }
        //Adapt the smartAI() some more if possible
        assessRepetitionPredictions();
    }
    
    /**
     * A method to help some portions of the smartAI() to adapt.
     */
    private void assessRepetitionPredictions()
    {
        //If we predicted the player's move based on repetition, and we fail 3 or more times, stop all repetition predictions.
        if(movePredictedRepetition && numOfRepetitionPredictions >= 3)
        {
            continueRepetitionPredictions = false;
        }
    }
    
    /**
     * A method that stops the game.
     */
    public void stopGame()
    {
        //Adjust some boolean variables
        gameStarted = false;
        gameStopped = true;
        
        //Remove all the buttons and actions from the world
        Playground.world.removeObjects(Playground.world.getObjects(Actions.class));
        Playground.world.removeObjects(Playground.world.getObjects(TextBox.class));
        Playground.world.removeObjects(Playground.world.getObjects(Buttons.class));
        
        //"Print" a message
        TextBox myFirstMessage = new TextBox("Thanks for playing!", 12, 9, 48, Color.BLACK, Playground.BACKGROUND_COLOR);
        TextBox mySecondMessage = new TextBox("You played: " + numOfGames + " game(s).", 12, 11, 24, Color.BLACK, Playground.BACKGROUND_COLOR);
        TextBox myThirdMessage = new TextBox("Click the \"Reset\" button to refresh the game.", 12, 12, 24, Color.BLACK, Playground.BACKGROUND_COLOR);
        
        //Stop Greenfoot's act method
        Greenfoot.stop();
    }
    
    /**
     * A method that restarts the game.
     */
    private void restart()
    {
        //Add a slight delay before clearing the playground
        Greenfoot.delay(800);
        
        //Clear the playground
        Playground.world.removeObjects(Playground.world.getObjects(Actions.class));
        Playground.world.removeObjects(Playground.world.getObjects(TextBox.class));
        
        //Adjust some boolean variables
        gameStarted = false;
        gameStopped = false;
        
        /**
         * Code to get the smartAI() to play against a truly random "player".
         */
        
        /**
        double randomNum = (Math.random() * 3);
        int roundedRandomNum = (int)(randomNum + 0.5);
        int castedRandomNum = (int)(randomNum);
        if(roundedRandomNum == 3)
        {
            roundedRandomNum--;
        }
        int averageNum = ((roundedRandomNum + castedRandomNum) / 2);
        if(averageNum == 0)
        {
            startGame(Playground.ROCK);
        }
        else if(averageNum == 1)
        {
            startGame(Playground.PAPER);
        }
        else if(averageNum == 2)
        {
            startGame(Playground.SCISSORS);
        }
        */
    }
    
    /**
     * @returns - Boolean that determines when it is safe to click buttons.
     */
    public boolean isButtonSafe()
    {
        //If the game is started, and not stopped. Then it is not safe to click buttons. Otherwise it is safe.
        if(gameStarted && !gameStopped)
        {
            return false;
        }
        else
        {
            return true;
        }
    }
    
    /**
     * Method that loads player moves from a specified URL.
     */
    private void loadPlayerMoves(URL myURL) throws FileNotFoundException
    {
        //Create a new BufferedReader
        BufferedReader myBufferedReader = null;
        try
        {
            myBufferedReader = new BufferedReader(new InputStreamReader(myURL.openStream())); //Initialize the BufferedReader
            String line = myBufferedReader.readLine(); //Read next line
            while (line != null) //While we have stuff to read, add it to some ArrayLists.
            {
                listOfPlayerMoves.add(Integer.parseInt(line));
                rollingListPlayerMoves.add(Integer.parseInt(line));
                line = myBufferedReader.readLine(); //Read the next line
            }
            myBufferedReader.close(); //Close the BufferedReader
        }
        catch(FileNotFoundException e)
        {
            throw e; //Throw the FileNotFoundException
        }
        catch(IOException e)
        {
            e.printStackTrace(); //Catch any other IOException and print it.
        }
    }
    
    /**
     * Some old code for saving and loading data from a text file.
     */
    
    /**
    private void save(String fileName) throws FileNotFoundException
    {
        PrintWriter myPrintWrtier = new PrintWriter(new FileOutputStream(fileName));
        for(int i = 7; i<listOfPlayerMoves.size(); i++)
        {
            myPrintWrtier.println(listOfPlayerMoves.get(i));
        }
        myPrintWrtier.println(-1);
        myPrintWrtier.close();
    }
    
    private void load(String fileName) throws FileNotFoundException
    {
        BufferedReader myBufferedReader = null;
        try {
            myBufferedReader = new BufferedReader(new FileReader(fileName));
            String line = myBufferedReader.readLine();
            while (line != null) {
                listOfPlayerMoves.add(Integer.parseInt(line));
                rollingListPlayerMoves.add(Integer.parseInt(line));
                line = myBufferedReader.readLine();
            }
            myBufferedReader.close();
        } catch(IOException e) {}
    }
    */
}